--
-- Simple Apple Game
-- by Marcus 'ReFreezed' Thunström
-- 2021-08-08
--
-- This example game has a level menu and a game world. Only the menu or the
-- world is "active" at a time. The world has the player character and a bunch
-- of apples on the ground. The goal is to walk over the apples to eat them.
-- When all apples are eaten the level is won.
--

-- Level data. When we start a level we use this data to create the world. We never
-- change anything here in-game - we just copy stuff from this when needed.
local LEVELS = {
	{id=1, name="The First Apple", apples={ {x=100,y=100} }},
	{id=2, name="Yummy Apples",    apples={ {x=100,y=100},{x=200,y=100} }},
	{id=3, name="More More More",  apples={ {x=100,y=100},{x=200,y=100},{x=100,y=200},{x=200,y=200} }},
}

-- Some settings.
local PLAYER_RADIUS     = 15
local PLAYER_MOVE_SPEED = 100 -- Units per second.
local APPLE_RADIUS      = 10

local menu  = nil -- We set this when we show a menu.
local world = nil -- We set this when we start a level.

local completedLevelIds = {} -- This will be saved to a savefile.

local menuFont = nil
local hudFont  = nil



local function saveProgress()
	print("Saving progress")

	-- The savefile will have level IDs separated by spaces.
	local dataString = table.concat(completedLevelIds, " ")

	love.filesystem.write("completedLevels.txt", dataString)
end

local function loadProgress()
	print("Loading progress")

	local dataString, err = love.filesystem.read("completedLevels.txt")
	if not dataString then
		-- The savefile probably doesn't exist.
		print(err)
		return
	end

	completedLevelIds = {} -- Make sure the array is empty.

	-- Match all the level IDs in the file data.
	for levelIdString in dataString:gmatch("%d+") do
		local levelId = tonumber(levelIdString)

		table.insert(completedLevelIds, levelId)
	end
end

local function resetProgress()
	print("Resetting progress")

	completedLevelIds = {}

	love.filesystem.remove("completedLevels.txt")
end



local function loadLevel(level)
	print("Starting level: "..level.name)
	menu = nil -- Close the menu, if one was open.

	-- Create world.
	world = {
		levelId = level.id,
		player  = {x=150, y=150},
		apples  = {},
	}

	-- Spawn apples.
	for _, appleCoords in ipairs(level.apples) do
		local apple = {x=appleCoords.x, y=appleCoords.y}
		table.insert(world.apples, apple)
	end
end

local function isLevelCompleted(levelId)
	-- Look if levelId is in the completedLevelIds array.
	for _, completedLevelId in ipairs(completedLevelIds) do
		if completedLevelId == levelId then
			return true
		end
	end

	return false -- The level is not completed.
end

local function markLevelAsCompleted(levelId)
	if isLevelCompleted(levelId) then
		return -- The level was already completed.
	end

	table.insert(completedLevelIds, levelId)
	saveProgress()
end



local function showLevelMenu()
	print("Showing level menu")
	world = nil -- Destroy the world, if one existed.

	menu = {
		buttons             = {},
		selectedButtonIndex = 1,
	}

	--
	-- Add buttons for the levels. Every button has a label and a callback
	-- function that we'll call when the button is pressed.
	--
	for i, level in ipairs(LEVELS) do
		local label = "Level " .. i .. ": " .. level.name

		if isLevelCompleted(level.id) then
			label = label .. " (COMPLETED)"
		end

		table.insert(menu.buttons, {
			label   = label,
			onPress = function()
				loadLevel(level)
			end,
		})
	end

	--
	-- Add a couple of extra buttons.
	--
	table.insert(menu.buttons, {
		label   = "Reset progress",
		onPress = function()
			resetProgress()

			-- Reset the menu so we see all levels as not completed.
			local i = menu.selectedButtonIndex -- Save the state of the old menu.
			showLevelMenu()
			menu.selectedButtonIndex = i -- Restore the state of the new menu.
		end,
	})

	table.insert(menu.buttons, {
		label   = "Exit",
		onPress = function()
			love.event.quit()
		end,
	})
end



function love.load()
	io.stdout:setvbuf("no") -- This makes any printing to the console happen immediately.
	io.stderr:setvbuf("no")

	menuFont = love.graphics.newFont(20)
	hudFont  = love.graphics.newFont(15)

	loadProgress()

	-- The first thing we show the player is the level menu.
	showLevelMenu()
end



function love.keypressed(key)
	--
	-- Menu controls.
	--
	if menu then
		if key == "down" then
			-- Select the next button.
			if menu.selectedButtonIndex < #menu.buttons then
				menu.selectedButtonIndex = menu.selectedButtonIndex + 1
			end

		elseif key == "up" then
			-- Select the previous button.
			if menu.selectedButtonIndex > 1 then
				menu.selectedButtonIndex = menu.selectedButtonIndex - 1
			end

		elseif key == "return" then
			-- Activate the selected button.
			menu.buttons[menu.selectedButtonIndex].onPress()
		end

	--
	-- World controls.
	--
	else
		if key == "escape" then
			-- Destroy the world and show the level menu.
			showLevelMenu()
		end
	end
end



function love.update(dt)
	if world then
		-- Move player.
		local player = world.player

		if love.keyboard.isDown("up")    then  player.y = player.y - PLAYER_MOVE_SPEED * dt  end
		if love.keyboard.isDown("down")  then  player.y = player.y + PLAYER_MOVE_SPEED * dt  end
		if love.keyboard.isDown("left")  then  player.x = player.x - PLAYER_MOVE_SPEED * dt  end
		if love.keyboard.isDown("right") then  player.x = player.x + PLAYER_MOVE_SPEED * dt  end

		-- Eat nearby apples.
		for i = #world.apples, 1, -1 do
			local apple    = world.apples[i]
			local dx       = player.x - apple.x
			local dy       = player.y - apple.y
			local distance = math.sqrt(dx*dx + dy*dy)

			if distance < PLAYER_RADIUS + APPLE_RADIUS then
				print("Ate apple")
				table.remove(world.apples, i)
			end
		end

		--
		-- If there are no more apples in the world we win the level and go
		-- back to the menu. We could choose to load the next level or
		-- something instead if we wanted to.
		--
		if #world.apples == 0 then
			print("Completed level")
			markLevelAsCompleted(world.levelId)
			showLevelMenu()
		end
	end
end



function love.draw()
	local LG = love.graphics

	--
	-- Draw menu.
	--
	if menu then
		local x = 20
		local y = 20

		LG.setFont(menuFont)

		-- Draw buttons.
		for i, button in ipairs(menu.buttons) do
			if i == menu.selectedButtonIndex then
				LG.setColor(0, 1, 0)
			else
				LG.setColor(.7, .7, .7)
			end
			LG.print(button.label, x, y)

			-- Draw the next button below this button.
			y = y + LG.getFont():getHeight() + 5
		end

	--
	-- Draw world.
	--
	else
		-- Draw Apples.
		for _, apple in ipairs(world.apples) do
			LG.setColor(1, 0, 0)
			LG.circle("fill", apple.x, apple.y, APPLE_RADIUS)
		end

		-- Draw Player.
		LG.setColor(.3, .3, 1)
		LG.circle("fill", world.player.x, world.player.y, PLAYER_RADIUS)

		-- Draw HUD.
		local text = "Apples remaining: " .. #world.apples
		LG.setFont(hudFont)
		LG.setColor(1, 1, 1)
		LG.print(text, 5, 5)
	end
end


